/* nespot.c - main source file for NESPOT Connecting

   Copyleft 2003, Gye-Ok Jung (hook7346@kornet.net)
   All rights reserved under GPL.
*/

#include <libnet.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pcap.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>


#include "B02_1x.h"


/* MD5 (Message Digest 5)
   Includings and Declarations, Functions
*/
#include "global.h"
#include "md5.h"

#ifndef MD
#define MD 5
#endif

static void MDString PROTO_LIST ((char *));
static void MDPrint PROTO_LIST ((unsigned char [16]));

#define MD5_CTX MD5_CTX
#define MDInit MD5Init
#define MDUpdate MD5Update
#define MDFinal MD5Final


// Transmission packet length == 802.1x Authentication packet total length ( Excluding Ethernet Header )
#define LIBNET_802_1x_H   0x32
#define B02_1x_AUTH_H     0x04
#define EAP_H             0x2e


// PCAP Receive Variable Values
#define PCAP_MAX_CNT 1
#define PCAP_SNAPSHOT 1024
#define PCAP_TIMEOUT 10

void packet_analyze(unsigned char *user, const struct pcap_pkthdr *h, const unsigned char *p);

char Filter_Rule[52] = "ether proto 0x888e and ether dst ";
char printable_eth_addr[18] = "::::::::::::::::";
u_char client_ether_addr[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
u_char ap_ether_addr[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

u_char challenge_request[17] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
u_char challenge_response[17];


void build_802_1x_Auth(struct B02_1x_auth_header *B02_1x_hdr, u_char *packet);
void build_EAP(struct B02_1x_auth_header *B02_1x_hdr, struct eap_header *eap_hdr, u_char *packet);
int receive_packet(char *device);


unsigned short recv_state_flag = 0;
char eap_id_number;

/* ׽ console main Լ  -> nespot_connect() Լ . GUI  ȣ. */
/* exit(-1)   Լ  return value  
 * GUI ʿ  ó */
/* nespot_connect( char *Nespot_id, char *Nespot_password, u_char *device ) */
int nespot_connect( char *Nespot_id, char *Nespot_password, u_char *device )
{
  register int i, j;
  int c, len;
  int packet_len;

  u_char *packet;

  /* u_char *device = NULL; */

  char err_buf[LIBNET_ERRBUF_SIZE];
  libnet_t *l;
  struct libnet_ether_addr *e;
  libnet_ptag_t data_tag, ether_tag;

  char eth_addr_802_1x[12];

  /*
  char *Nespot_id = "";
  char *Nespot_password = "";
  */
  
  u_char challenge_tmp_buf[50];

  struct B02_1x_auth_header B02_1x_hdr;
  struct eap_header eap_hdr;
  struct B02_1x_auth_header *B02_1x_hdr_ptr = &B02_1x_hdr;
  struct eap_header *eap_hdr_ptr = &eap_hdr;
  
  /* α׷   κ   -> GUI   id,pass,device  ѾԳ Ȯ */
  if( strlen(Nespot_id) < 1 )
  {
	  fprintf(stderr, "Name was not entered.\n");
	  return -1;
  } else if( strlen(Nespot_password) < 1 ) {
	  fprintf(stderr, "Password was not entered\n" );
	  return -1;
  } else if( strlen(device) < 1 ) {
	  fprintf(stderr, "Device was not selected\n" );
	  return -1;
  }

  /* (Libnet) Initialize */

  l = libnet_init(LIBNET_LINK_ADV, device, err_buf);
  if(l == NULL) {
    fprintf(stderr, "libnet_init() Error : %s\n", err_buf);
	return -1; /* exit(EXIT_FAILURE); */
  }


  /* Get the own Wireless NIC Ethernet Address */
  e = libnet_get_hwaddr(l);
  if(e == NULL) {
    fprintf(stderr, "libnet_get_hwaddr() Error : %s\n", libnet_geterror(l));
    return -1; /* exit(EXIT_FAILURE); */
  }

  /* 802.1x Wireless NIC Ethernet Address print */
  for(i=0; i<6; i++) {
    fprintf(stdout, "%02X ", e->ether_addr_octet[i]);

    client_ether_addr[i] = e->ether_addr_octet[i]; /* Client Ethernet Address Assignment : own Wireless NIC MAC */
  }
  fprintf(stdout, "\n");

  for(i=0; i<6; i++) { /* Convert 802.1x Wireless NIC Ethernet Address to ASCII Readable */
    sprintf(eth_addr_802_1x+(2*i), "%02X", e->ether_addr_octet[i]);
  }

  /* Make String for Filter Construction */
  for(i=0; i<6; i++) {
    for(j=0; j<2; j++) {
      printable_eth_addr[j+(i*3)]  = eth_addr_802_1x[j+(i*2)];
    }
  }
  printf("Wireless NIC MAC -> %s\n", printable_eth_addr);



  /* Packet Filtering Rule Construction */
  strncat(Filter_Rule, printable_eth_addr, 17);
  printf("Filter Rule --> %s\n", Filter_Rule);



  /* EAPOL start Packet Construction */
  B02_1x_hdr_ptr->auth_version = 0x01;
  B02_1x_hdr_ptr->auth_type = 0x01;
  B02_1x_hdr_ptr->auth_length = htons(0x0000);

  packet_len = LIBNET_802_1x_H;

  packet = (unsigned char *) malloc(packet_len);
  if(packet == NULL) {
    perror("malloc");
    libnet_destroy(l);
    return -1; /* exit(-1); */
  }

  build_802_1x_Auth(B02_1x_hdr_ptr, packet);


  data_tag = LIBNET_PTAG_INITIALIZER;
  data_tag = libnet_build_data(packet, packet_len, l, data_tag);
  if(data_tag == -1) {
    (void) fprintf(stderr, "Can't build data block: %s \n", libnet_geterror(l));
    libnet_destroy(l);
    free(packet);
    return -1; /* exit(-1); */
  }

  ether_tag = LIBNET_PTAG_INITIALIZER;
  ether_tag = libnet_build_ethernet(ap_ether_addr, client_ether_addr, 0x888e, NULL, 0, l, ether_tag);
  if(ether_tag == -1) {
    (void) fprintf(stderr, "Can't build ethernet header block: %s \n", libnet_geterror(l));
    libnet_destroy(l);
    free(packet);
    return -1; /* exit(-1); */
  }

  /* EAPOL start Packet Transmission */
  len = libnet_write(l);
  if(len == -1) {
    (void) fprintf(stderr, "Write Error: %s \n", libnet_geterror(l));
    libnet_destroy(l);
    free(packet);
    return -1; /* exit(-1); */
  }
  fprintf(stderr, "libnet write succeeded!! EAPOL packet transmission Completed!! \n");

  free(packet);


  /* EAP Identity Packet Receive [RFC2284] : This must be the first receiving packet */
  do {
    receive_packet(device);

  }while(recv_state_flag == 0);
  

  /* EAP resopnse Packet Construction */
  B02_1x_hdr_ptr->auth_version = 0x01;
  B02_1x_hdr_ptr->auth_type = 0x00;
  B02_1x_hdr_ptr->auth_length = htons(0x0005 + strlen(Nespot_id));

  eap_hdr_ptr->eap_code = 0x02;
  eap_hdr_ptr->eap_id = eap_id_number;
  eap_hdr_ptr->eap_length = B02_1x_hdr_ptr->auth_length;
  eap_hdr_ptr->eap_type = 0x01;
  strncpy(eap_hdr_ptr->eap_data, Nespot_id, strlen(Nespot_id));

  packet_len = LIBNET_802_1x_H;

  packet = (unsigned char *) malloc(packet_len);
  if(packet == NULL) {
    perror("malloc");
    libnet_destroy(l);
    return -1; /* exit(-1); */
  }

  build_EAP(B02_1x_hdr_ptr, eap_hdr_ptr, packet);

  data_tag = LIBNET_PTAG_INITIALIZER;
  data_tag = libnet_build_data(packet, packet_len, l, data_tag);
  if(data_tag == -1) {
    (void) fprintf(stderr, "Can't build data block: %s \n", libnet_geterror(l));
    libnet_destroy(l);
    free(packet);
    return -1; /* exit(-1); */
  }

  ether_tag = LIBNET_PTAG_INITIALIZER;
  ether_tag = libnet_build_ethernet(ap_ether_addr, client_ether_addr, 0x888e, NULL, 0, l, ether_tag);
  if(ether_tag == -1) {
    (void) fprintf(stderr, "Can't build ethernet header block: %s \n", libnet_geterror(l));
    libnet_destroy(l);
    free(packet);
    return -1; /* exit(-1); */
  }

  /* EAP resopnse Packet Transmission */
  len = libnet_write(l);
  if(len == -1) {
    (void) fprintf(stderr, "Write Error: %s \n", libnet_geterror(l));
    libnet_destroy(l);
    free(packet);
    return -1; /* exit(-1); */
  }
  fprintf(stderr, "libnet write succeeded!! EAP Response packet (Identity) transmission Completed!! \n");

  free(packet);


  /* MD5 Challeng request Packet Receive [RFC2284] : This must be the second receiving packet */
  do {
    receive_packet(device);

  }while(recv_state_flag == 1);


  /* MD5 Checksum Input Buffer Making */
  challenge_tmp_buf[0] = eap_id_number;
  strncpy(challenge_tmp_buf+1, Nespot_password, strlen(Nespot_password)+1);
  strncat(challenge_tmp_buf  , challenge_request, 16); 

  MDString(challenge_tmp_buf);

  /* Challenge Response Data Making */
  challenge_tmp_buf[0] = 0x10;
  strncpy(challenge_tmp_buf+1, challenge_response, 17);
  strncat(challenge_tmp_buf  , Nespot_id, strlen(Nespot_id));

  /* EAP resopnse Packet Construction */
  B02_1x_hdr_ptr->auth_version = 0x01;
  B02_1x_hdr_ptr->auth_type = 0x00;
  B02_1x_hdr_ptr->auth_length = htons(0x0005 + strlen(challenge_tmp_buf));

  eap_hdr_ptr->eap_code = 0x02;
  eap_hdr_ptr->eap_id = eap_id_number;
  eap_hdr_ptr->eap_length = B02_1x_hdr_ptr->auth_length;
  eap_hdr_ptr->eap_type = 0x04;
  strncpy(eap_hdr_ptr->eap_data, challenge_tmp_buf, strlen(challenge_tmp_buf));

  packet_len = LIBNET_802_1x_H;

  packet = (unsigned char *) malloc(packet_len);
  if(packet == NULL) {
    perror("malloc");
    libnet_destroy(l);
    return -1; /* exit(-1); */
  }

  build_EAP(B02_1x_hdr_ptr, eap_hdr_ptr, packet);

  data_tag = LIBNET_PTAG_INITIALIZER;
  data_tag = libnet_build_data(packet, packet_len, l, data_tag);
  if(data_tag == -1) {
    (void) fprintf(stderr, "Can't build data block: %s \n", libnet_geterror(l));
    libnet_destroy(l);
    free(packet);
    return -1; /* exit(-1); */
  }

  ether_tag = LIBNET_PTAG_INITIALIZER;
  ether_tag = libnet_build_ethernet(ap_ether_addr, client_ether_addr, 0x888e, NULL, 0, l, ether_tag);
  if(ether_tag == -1) {
    (void) fprintf(stderr, "Can't build ethernet header block: %s \n", libnet_geterror(l));
    libnet_destroy(l);
    free(packet);
    return -1; /* exit(-1); */
  }

  /* EAP resopnse Packet Transmission */
  len = libnet_write(l);
  if(len == -1) {
    (void) fprintf(stderr, "Write Error: %s \n", libnet_geterror(l));
    libnet_destroy(l);
    free(packet);
    return -1; /* exit(-1); */
  }
  fprintf(stderr, "libnet write succeeded!! EAP Response packet(MD5 Challenge) transmission Completed!! \n");

  libnet_destroy(l);
  free(packet);

  /* MD5 Challeng request Packet Receive [RFC2284] : This must be the second receiving packet */
  do {
    receive_packet(device);

  }while(recv_state_flag == 2);


  return 0; /* return(0); */

}
/* Nespot console main Լ  -> nespot_connect() Լ . GUI  ȣ */

void build_802_1x_Auth(struct B02_1x_auth_header *B02_1x_hdr, u_char *packet)
{
  register int i;
  u_char *p = packet;

  u_char *h = (u_char *)B02_1x_hdr;

  for(i=0; i<B02_1x_AUTH_H; i++)
    *p++ = *h++;
}

void build_EAP(struct B02_1x_auth_header *B02_1x_hdr, struct eap_header *eap_hdr, u_char *packet)
{
  register int i;
  u_char *p = packet;

  u_char *h = (u_char *)B02_1x_hdr;

  for(i=0; i<B02_1x_AUTH_H; i++)
    *p++ = *h++;

  h = (u_char *)eap_hdr;

  for(i=0; i<EAP_H; i++)
    *p++ = *h++;
}


int receive_packet(char *device)
{
  char errbuf[PCAP_ERRBUF_SIZE];
  pcap_t *pd;
  struct bpf_program fcode;
  //  bpf_u_int32 net;
  bpf_u_int32 netmask = 0xffffff00;


  //pcap device open for packet capturing
  if((pd = pcap_open_live(device, PCAP_SNAPSHOT, 1, PCAP_TIMEOUT, errbuf)) == NULL) {
    perror(errbuf);
    return -1; /* exit(1); */
  }
 
  if(pcap_compile(pd, &fcode, Filter_Rule, 0, netmask) < 0) {
    perror(pcap_geterr(pd));
    return -1; /* exit(1); */
  }
  // set filtering rule
  if(pcap_setfilter(pd, &fcode) < 0) {
    perror(pcap_geterr(pd));
    return -1; /* exit(1); */
  }
  // read packet forever
  if(pcap_loop(pd, PCAP_MAX_CNT, packet_analyze, 0) < 0) {
    perror(pcap_geterr(pd));
    return -1; /* exit(1); */
  }
  
  pcap_close(pd);

  return 0;
}

/*
 * packet_analyze function
 * identify and print packets
 */
void packet_analyze(unsigned char *user, const struct pcap_pkthdr *h, const unsigned char *packet_offset)
{

  static int count = 0; // count for packet numbering
  int len;

  char buf[EAP_DATA_LENGTH];
  register int i, j, k;

  struct ether_header *ether_hdr_instance;
  struct B02_1x_auth_header *B02_1x_hdr_instance;
  struct eap_header *eap_hdr_instance;

  ether_hdr_instance = (struct ether_header *)packet_offset;
  packet_offset += sizeof(struct ether_header);

  B02_1x_hdr_instance = (struct B02_1x_auth_header *)packet_offset;
  packet_offset += sizeof(struct B02_1x_auth_header);

  eap_hdr_instance = (struct eap_header *)packet_offset;
  packet_offset += 0x05;
  //  packet_offset += sizeof(struct eap_header);

  /* Captured Packet Count */
  printf("PACKET %d\n", count++);
  printf("--------------------------------------------------------------------\n");

  /* In Captured Packet, Ethernet Destination MAC addresss */
  printf("ethernet destnation address : \n\t\t");
  for(i=0; i < ETH_ALEN; i++) {
    printf("%02X ", ether_hdr_instance->ether_dhost[i]);
  }
  printf("\n");

  /* In Captured Packet, Ethernet Source MAC addresss */
  printf("ethernet source address : \n\t\t");
  for(i=0; i < ETH_ALEN; i++) {
    printf("%02X ", ether_hdr_instance->ether_shost[i]);
  }
  printf("\n");

  printf("ether_type : 0x%04x\n", ntohs(ether_hdr_instance->ether_type));
  printf("\n");


  /* 802.1x Authentication Header Print */
  printf("802.1x Auth. -> version : %d\n", B02_1x_hdr_instance->auth_version);
  printf("802.1x Auth. -> type    : %d\n", B02_1x_hdr_instance->auth_type);
  printf("802.1x Auth. -> length  : %d\n", ntohs(B02_1x_hdr_instance->auth_length));

  len = ntohs(B02_1x_hdr_instance->auth_length);
  printf("\n");


  if( eap_hdr_instance->eap_code == 1 ) {
    
    /* EAP Header Print */
    printf("EAP -> code   : %d\n", eap_hdr_instance->eap_code);
    printf("EAP -> id     : %d\n", eap_hdr_instance->eap_id);
    printf("EAP -> length : %d\n", ntohs(eap_hdr_instance->eap_length));
    printf("EAP -> type   : %d\n", eap_hdr_instance->eap_type);
    
    /* EAP Header -> Data Print : be variable according to the eap_type */
    if((eap_hdr_instance->eap_type) == 1) {
      
      strncpy(buf, eap_hdr_instance->eap_data, len-5);
      buf[len-5] = '\0';
      
      printf("EAP -> DATA : %s\n", buf);
    }
    else if((eap_hdr_instance->eap_type) == 4) {
      
      i = len - 5;
      j = 0;
      
      printf("Challenge request Value --------->\n");
      strncpy(challenge_request, packet_offset+1, 16);
      for (k=0; k<17; k++) 
	{
	  printf("%02x ", challenge_request[k]);
	}
      printf("\n\n");
      
      printf("DATA --------->\n");
      while((i--) > 0) {
	printf("%02X ", *(packet_offset++));
	if(!(++j % 16))
	  printf("\n");
      }
      printf("\n");
      
    }
  }
  else if((eap_hdr_instance->eap_code) == 3) {
    
    printf("Nespot Connection Succeeded!! Good Luck!!\n\n");
  }
  else if((eap_hdr_instance->eap_code) == 4) {
    
    printf("Nespot Connection Failure!!\n\n");
  }
  else {
    printf("Unknown EAP code !! Please, Dubugging\n");
  }
  
  printf("--------------------------------------------------------------------\n\n");

  if ( recv_state_flag == 0 ) {
    if ( eap_hdr_instance->eap_code == eap_hdr_instance->eap_type ) { // eap_code == 1 and eap_type == 1

      eap_id_number = eap_hdr_instance->eap_id;
      for(i=0; i < 6; i++) {
	ap_ether_addr[i] = ether_hdr_instance->ether_shost[i];
      }

      recv_state_flag++; // Wanted Packet Accepted !!
    }
  }
  else if ( recv_state_flag == 1) {
    if ( eap_hdr_instance->eap_code == (eap_hdr_instance->eap_type - 3) ) { // eap_code == 1 and eap_type == 4

      eap_id_number = eap_hdr_instance->eap_id;

      recv_state_flag++; // Wanted Packet Accepted !!
    }
  }
  else if ( recv_state_flag == 2) {
    if ( eap_hdr_instance->eap_code == 3 || eap_hdr_instance->eap_code == 4 ) { // eap_code == 3 or eap_code == 4
      recv_state_flag++; // Wanted Packet Accepted !!
    }
  }


  return ;
}


/* Digests a string and prints the result.
 */
static void MDString (string)
char *string;
{
  MD5_CTX context;
  unsigned char digest[16];
  unsigned int len = strlen (string);

  MDInit (&context);
  MDUpdate (&context, string, len);
  MDFinal (digest, &context);

  printf ("MD%d (\"%s\") = ", MD, string);
  MDPrint (digest);
  printf ("\n");
}

/* Prints a message digest in hexadecimal.
 */
static void MDPrint (digest)
unsigned char digest[16];
{

  unsigned int i;

  for (i = 0; i < 16; i++) {
     printf ("%02x", digest[i]);
     sprintf (challenge_response+i,"%c", digest[i]);
  }
  challenge_response[16] = '\0';
}
